// M5AtomS3 LCD Test
#include <SPI.h>

#include "image1.h"
#include "image2.h"
#include "image3.h"
#include "font8x16.h"

#define LCD_CLK 17
#define LCD_MOSI 21
#define LCD_RS 33
#define LCD_CS 15
#define LCD_RST 34
#define LCD_BL 16

#define LCD_SLEEP_OUT 0x11    // SLEEP OUT
#define LCD_PTLON 0x12        // Partial Mode ON
#define LCD_NORON 0x13        // Normal Display Mode ON
#define LCD_INVOFF 0x20       // Display Inversion OFF
#define LCD_INVON 0x21        // Display Inversion ON
#define LCD_DISPLAY_OFF 0x28  // DISPLAY OFF
#define LCD_DISPLAY_ON 0x29   // DISPLAY ON
#define LCD_COLUMN_ADDRESS_SET 0x2a  // COL SET
#define LCD_PAGE_ADDRESS_SET 0x2b    // ROW SET
#define LCD_MEMORY_WRITE 0x2c        // MEMORY WRITE
#define LCD_PARTIAL_AREA 0x30        // Partial Area
#define LCD_VSCROLL_DEF  0x33        // Vertical Scrolling Definition
#define LCD_VSCROLL_ADDR 0x37        // Vertical Scrolling Start Address

#define LCD_MEMORY_ACCESS_CONTROL 0x36  // 表示向き
#define LCD_COLMOD 0x3a                 // 表示モード

#define LCD_COLOR16 0x55      // 1Dot/16Bitモード
#define LCD_COLOR18 0x66      // 1Dot/18Bitモード

#define LCD_WIDTH 128
#define LCD_HEIGHT 128

#define LCD_OFFSET_COL 2
#define LCD_OFFSET_ROW 1

#define HBYTE(u) ((u >> 8) & 0xFF)
#define LBYTE(u) (u & 0xFF)

#ifndef _swap_int16_t
#define _swap_int16_t(a, b) { int16_t t = a ; a = b; b = t; }
#endif

uint8_t SPIBuf[512] ;     // SPI転送用バッファ
uint8_t fontColorF[2] ;   // TEXT FOR GROND COLOR
uint8_t fontColorB[2] ;   // TEXT BACK GROND COLOR

// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
// LCDにコマンドを送信
void lcdSendCommand(uint8_t command) {
  digitalWrite(LCD_CS, LOW); // LCDセレクト
  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(LCD_CS, HIGH);  // LCDセレクト解放
}

// LCDセレクトにコマンド+1バイトデータを送信
void lcdSendCommand(uint8_t command, uint8_t data1) {
  digitalWrite(LCD_CS, LOW); // LCDセレクト
  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(LCD_RS, HIGH); // Command mode
  SPI.transfer(data1);
  digitalWrite(LCD_CS, HIGH);  // LCDセレクト解放
}

void screenDir(int dir) {
  //  MY - 0x80 // Row Address Order
  //  MX - 0x40 // Column Address Order
  //  MV - 0x20 // Row / Column Exchange
  //  ML - 0x10 // Vertical Refresh Order
  //  MRGB - 0x08 // RBG - BGR
  //  MH - 0x04 // Horizontal Refresh ORDER
  uint8_t PR = 0 ;
  switch(dir) {
    case 1:
      PR = 0x08 ;   // MRGB           // DIR-1 : 縦
      break ;
    case 3:
      PR = 0xC8;    // MX | MY | MRGB // DIR-3 : 縦
      break ;
    case 2:
      PR = 0x68 ;   // MX | MV | MRGB //　dir-2 : 横
      break ;
    case 4:
      PR = 0xA8 ;   // MY | MV | MRGB //　dir-4 : 横
      break ;
  }
  lcdSendCommand(LCD_MEMORY_ACCESS_CONTROL,PR) ;
}

// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
void lcd_init() {
  // --- HARDWARE Reset
  digitalWrite(LCD_RST, LOW);
  delay(120);
  digitalWrite(LCD_RST, HIGH);
  delay(120);    // このdelayは必要

  // --- LCD Wakeup
  lcdSendCommand(LCD_SLEEP_OUT) ;
  delay(120);    // このdelayは必要

  lcdSendCommand(LCD_NORON) ; //  Normal Display Mode ON
  lcdSendCommand(LCD_INVON) ; //  Display Inversion ON

  // ----- 動作モード設定 -----
  screenDir(3) ;
  lcdSendCommand(0x35,0x00) ;

  // ピクセルフォーマットを1Dot/16bitに設定
  lcdSendCommand(LCD_COLMOD,LCD_COLOR16);

  fillScreen(0xFF) ;

  // 画面表示ON
  lcdSendCommand(LCD_DISPLAY_ON);
}

// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
void setScroolArea(uint8_t TFA , uint8_t VSA) {
  uint8_t BFA = 128 - TFA - VSA ;
  digitalWrite(LCD_CS, LOW);  //通信開始

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_VSCROLL_DEF);
  digitalWrite(LCD_RS, HIGH); // Command mode
  SPI.transfer(0);
  SPI.transfer(TFA);
  SPI.transfer(0);
  SPI.transfer(VSA);
  SPI.transfer(0);
  SPI.transfer(BFA);
  digitalWrite(LCD_RS, LOW); // Command mode
  digitalWrite(LCD_CS, HIGH); //通信終了
}

void setScroolAddr(uint8_t ADDR) {
  digitalWrite(LCD_CS, LOW);  //通信開始

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_VSCROLL_ADDR);
  digitalWrite(LCD_RS, HIGH); // Command mode
  SPI.transfer(0);
  SPI.transfer(ADDR);
  digitalWrite(LCD_RS, LOW); // Command mode
  digitalWrite(LCD_CS, HIGH); //通信終了
}

// BITマップを表示する::基本fillと同じ作り
void dispBitMap(uint16_t startX , uint16_t startY , uint16_t width, uint16_t hight, uint8_t bitMap[]) {
  uint16_t endX = startX + width - 1;
  uint16_t endY = startY + hight - 1;

  digitalWrite(LCD_CS, LOW);  //通信開始

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_COLUMN_ADDRESS_SET);
  digitalWrite(LCD_RS, HIGH); // Command mode
  SPI.transfer16(startX+LCD_OFFSET_COL);
  SPI.transfer16(endX+LCD_OFFSET_COL);

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_PAGE_ADDRESS_SET);
  digitalWrite(LCD_RS, HIGH); // Command mode
  SPI.transfer16(startY+LCD_OFFSET_ROW);
  SPI.transfer16(endY+LCD_OFFSET_ROW);

  int pos = 0 ;
  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_MEMORY_WRITE);
  digitalWrite(LCD_RS, HIGH); // Command mode
  for (int row = 0 ; row < hight ; row ++) {
    for (int col = 0 ; col < width * 2 ; col ++ ) {
      SPIBuf[col] = bitMap[pos];
      pos ++ ;
    }
    SPI.transfer(SPIBuf, width * 2);
  }
  digitalWrite(LCD_RS, LOW); // Command mode
  digitalWrite(LCD_CS, HIGH); //通信終了
}

// △▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽
void fillRect(uint16_t x, uint16_t y, uint16_t width, uint16_t hight, uint16_t color) {
  long num = (long)hight * (long)width ;
  uint16_t endX = x + width - 1;
  uint16_t endY = y + hight - 1;

  digitalWrite(LCD_CS, LOW);  //通信開始

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_COLUMN_ADDRESS_SET);
  digitalWrite(LCD_RS, HIGH); // data mode
  SPI.transfer16(x+LCD_OFFSET_COL);
  SPI.transfer16(endX+LCD_OFFSET_COL);

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_PAGE_ADDRESS_SET);
  digitalWrite(LCD_RS, HIGH); // data mode
  SPI.transfer16(y+LCD_OFFSET_ROW);
  SPI.transfer16(endY+LCD_OFFSET_ROW);

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_MEMORY_WRITE);
  digitalWrite(LCD_RS, HIGH); // data mode
  for (long i = 0 ; i < num ; i++) {
    SPI.transfer16(color);
  }

  digitalWrite(LCD_RS, LOW); // Command mode
  digitalWrite(LCD_CS, HIGH); //通信終了
}

void fill(uint16_t xs,uint16_t ys,uint16_t xe,uint16_t ye,uint16_t color) {
  uint16_t hight = ye - ys + 1 ;
  uint16_t width = xe - xs + 1 ;
  fillRect(xs,ys,width,hight,color) ;
}

void fillScreen(uint16_t color) {
  fill(0,0,LCD_WIDTH-1,LCD_HEIGHT-1,color) ;
}

// △▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽
void dispStr(char str[] , uint8_t x , uint8_t y) {
  for (int i=0;str[i] != 0 ; i++) {
      dispFont(str[i],x,y) ;
      x += 8 ;
      if (x+8>=LCD_WIDTH) {
        x = 0 ;
        y += 16 ;
        if (y + 16 >= LCD_HEIGHT) {
          y = 0 ; 
        }
      }
  }
}

// △▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽
void dispFont(uint8_t code , uint8_t x, uint8_t y) {
  int pos = (code - 0x20) * 16 ;
  memset(SPIBuf,0,8*16*2) ;
  for (int i=0;i<16;i++) {
    uint8_t data = lcd_font8x16[pos + i];
    uint8_t bitMast = 0x01 ; 
    for (int j=0;j<8;j++) {
      int fpos = 0 ;
      if (i<8) {
        fpos = (j*8+i)*2 ;
      } else {
        fpos = ((j+8)*8+(i-8))*2 ;
      }
      if ((data & bitMast) == 0) {
        SPIBuf[fpos] = fontColorB[0] ;
        SPIBuf[fpos+1] = fontColorB[1] ;
      } else {
        SPIBuf[fpos] = fontColorF[0] ;
        SPIBuf[fpos+1] = fontColorF[1] ;
      }
      bitMast <<= 1 ;
    }
  }

  uint8_t endX = x + 7 ;
  uint8_t endY = y + 15;

  digitalWrite(LCD_CS, LOW);  //通信開始

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_COLUMN_ADDRESS_SET);
  digitalWrite(LCD_RS, HIGH); // data mode
  SPI.transfer16(x+LCD_OFFSET_COL);
  SPI.transfer16(endX+LCD_OFFSET_COL);

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_PAGE_ADDRESS_SET);
  digitalWrite(LCD_RS, HIGH); // data mode
  SPI.transfer16(y+LCD_OFFSET_ROW);
  SPI.transfer16(endY+LCD_OFFSET_ROW);

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_MEMORY_WRITE);
  digitalWrite(LCD_RS, HIGH); // data mode
  SPI.transfer(SPIBuf, 8*16*2);
  digitalWrite(LCD_RS, LOW); // Command mode
  digitalWrite(LCD_CS, HIGH); //通信終了
}

// ▼▼▼▼▼▼ Adafruit_GFX_Libraryからパクった描画関数
// ローテーションなし、Windowクリップなし、描画フレームへなしにして不要部分を削ってます

// 1Dot描画
void drawPixel(uint16_t x, uint16_t y, uint16_t color) {
  if (x < 0 || x >= LCD_WIDTH || y < 0 || y >= LCD_HEIGHT) return ;
  uint16_t xe = x ;
  uint16_t ye = y ;

  digitalWrite(LCD_CS, LOW); // TFTセレクト
  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_COLUMN_ADDRESS_SET);
  digitalWrite(LCD_RS, HIGH); // data mode
  SPI.transfer16(x+LCD_OFFSET_COL);
  SPI.transfer16(xe+LCD_OFFSET_COL);

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_PAGE_ADDRESS_SET);
  digitalWrite(LCD_RS, HIGH); // datamand mode
  SPI.transfer16(y+LCD_OFFSET_ROW);
  SPI.transfer16(ye+LCD_OFFSET_ROW);

  digitalWrite(LCD_RS, LOW); // Command mode
  SPI.transfer(LCD_MEMORY_WRITE);
  digitalWrite(LCD_RS, HIGH); // data mode
  SPI.transfer16(color);

  digitalWrite(LCD_RS, LOW); // Command mode
  digitalWrite(LCD_CS, HIGH); // TFT解放
}

// Line描画
void drawLine(int x0, int y0, int x1, int y1, uint16_t color) {
  int16_t steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep) {
    _swap_int16_t(x0, y0);
    _swap_int16_t(x1, y1);
  }
  if (x0 > x1) {
    _swap_int16_t(x0, x1);
    _swap_int16_t(y0, y1);
  }

  int16_t dx, dy;
  dx = x1 - x0;
  dy = abs(y1 - y0);

  int16_t err = dx / 2;
  int16_t ystep;

  if (y0 < y1) {
    ystep = 1;
  } else {
    ystep = -1;
  }

  for (; x0 <= x1; x0++) {
    if (steep) {
      drawPixel(y0, x0, color);
    } else {
      drawPixel(x0, y0, color);
    }
    err -= dy;
    if (err < 0) {
      y0 += ystep;
      err += dx;
    }
  }
}

// 矩形
void drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) {
  int xe = x + w - 1 ;
  int ye = y + h - 1 ;
  drawLine(x, y, xe,y ,color);
  drawLine(x, y, x ,ye,color);
  drawLine(xe,y, xe,ye,color);
  drawLine(x ,ye,xe,ye,color);
}

// 円
void drawCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;

  drawPixel(x0, y0 + r, color);
  drawPixel(x0, y0 - r, color);
  drawPixel(x0 + r, y0, color);
  drawPixel(x0 - r, y0, color);

  while (x < y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;

    drawPixel(x0 + x, y0 + y, color);
    drawPixel(x0 - x, y0 + y, color);
    drawPixel(x0 + x, y0 - y, color);
    drawPixel(x0 - x, y0 - y, color);
    drawPixel(x0 + y, y0 + x, color);
    drawPixel(x0 - y, y0 + x, color);
    drawPixel(x0 + y, y0 - x, color);
    drawPixel(x0 - y, y0 - x, color);
  }
}

// 塗りつぶ円
void fillCircle(int16_t x0, int16_t y0, int16_t r,uint16_t color) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;
  int16_t px = x;
  int16_t py = y;
  int16_t delta = 1 ;
  
  fillRect(x0, y0 - r , 1 , r * 2 , color);

  while (x < y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;
    if (x < (y + 1)) {
        fillRect(x0 + x , y0 - y , 1 , y * 2 + delta -1 , color);
        fillRect(x0 - x , y0 - y , 1 , y * 2 + delta -1 , color);
    }
    if (y != py) {
        fillRect(x0 + py , y0 - px , 1 , px * 2 + delta -1 , color);
        fillRect(x0 - py , y0 - px , 1 , px * 2 + delta -1 , color);
      py = y;
    }
    px = x;
  }
}


// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
// SPISettings settings = SPISettings(8000000, MSBFIRST, SPI_MODE0);

void setup() {
  Serial.begin(9600);
  delay(2000) ;
  Serial.println(F("GC9107 Test"));
  Serial.println() ;
  Serial.println() ;

  // ---- SPI
  SPI.begin();  //SPIを初期化、SCK、MOSI、SSの各ピンの動作は出力、SCK、MOSIはLOW、SSはHIGH
  SPI.setClockDivider(SPI_CLOCK_DIV2);
  SPI.setBitOrder(MSBFIRST);
  SPI.setDataMode(SPI_MODE0);

  // --- Back Light
  analogWrite(LCD_BL,255) ; // PWM 出力で明るさを調整出来る
  // -----  LCD INITIAL -----
  pinMode(LCD_CS, OUTPUT);
  pinMode(LCD_RS, OUTPUT);
  pinMode(LCD_RST, OUTPUT);
  digitalWrite(LCD_CS,HIGH) ;
  digitalWrite(LCD_RS,HIGH) ;
  digitalWrite(LCD_RST,HIGH) ;

  lcd_init() ;
}

// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
void loop() {
  // ----- BITMAP IMAGE -----
  setScroolArea(0,128) ;
  setScroolAddr(0) ;
  dispBitMap(0 , 0 , 128, 128, image_data_image3) ;
  delay(2000) ;
  setScroolArea(32,64) ;
  for (int j=0;j<3;j++) {
    for (int i=32;i<=100;i+=2) {
      setScroolAddr(i) ;
      delay(50) ;
    }
  }
  delay(2000) ;
  setScroolArea(0,128) ;
  setScroolAddr(0) ;

  dispBitMap(0 , 0 , 128, 128, image_data_image1) ;
  delay(2000) ;
  dispBitMap(0 , 0 , 128, 128, image_data_image2) ;
  delay(2000) ;

  // ----- TEXT -----
  fillScreen(0xFFFF) ;  // B11111 111 111 11111 :  WHITE
  fontColorB[0] = 0xFF ; fontColorB[1] = 0xFF ; // TEXT BACK GROND COLOR
  fontColorF[0] = 0xF8 ; fontColorF[1] = 0x00 ; // TEXT FOR GROND COLOR
  dispStr("ABCDEFGHIJ",0,0) ;
  fontColorF[0] = 0xF8 ; fontColorF[1] = 0x00 ; // TEXT FOR GROND COLOR
  dispStr("0123456789",0,16) ;
  fontColorF[0] = 0x07 ; fontColorF[1] = 0x80 ; // TEXT FOR GROND COLOR
  dispStr("ABCDEFGHIJ",0,32) ;
  fontColorF[0] = 0x07 ; fontColorF[1] = 0x80 ; // TEXT FOR GROND COLOR
  dispStr("0123456789",0,48) ;
  fontColorF[0] = 0x00 ; fontColorF[1] = 0x1F ; // TEXT FOR GROND COLOR
  dispStr("ABCDEFGHIJ",0,64) ;
  fontColorF[0] = 0x00 ; fontColorF[1] = 0x1F ; // TEXT FOR GROND COLOR
  dispStr("0123456789",0,80) ;
  delay(2000) ;
  
  int xs , ys , xe , ye , w , h ;
  int c ;

  // -- LINE
  fillScreen(0x0000) ; 

  xs = 0 ;  ys = 0 ;  xe = 0 ;  ye = 127 ;
  for (int i=0;i<8;i++) {
    drawLine(xs,ys,xe,ye,0xF800) ;
    ys += 16 ;
    xe += 16 ;
  }

  xs = 0 ;  ys = 127 ;  xe = 127 ;  ye = 127 ;
  for (int i=0;i<8;i++) {
    drawLine(xs,ys,xe,ye,0x07E0) ;
    xs += 16 ;
    ye -= 16 ;
  }
  
  xs = 0 ;  ys = 0 ;  xe = 127 ;  ye = 0 ;
  for (int i=0;i<8;i++) {
    drawLine(xs,ys,xe,ye,0x001F) ;
    xs += 16 ;
    ye += 16 ;
  }
  
  xs = 0 ;  ys = 127 ;  xe = 0 ;  ye = 0 ;
  for (int i=0;i<8;i++) {
    drawLine(xs,ys,xe,ye,0xFFFF) ;
    ys -= 16 ;
    xe += 16 ;
  }
  delay(2000);

  // -- rect
  fillScreen(0x0000) ; 
  xs = 0 ;  ys = 0 ;  w = 128 ;  h = 128 ;
  for (int i=0;i<8;i++) {
    switch (i % 3) {
      case 0: drawRect(xs,ys,w,h,0xF800) ; break ;
      case 1: drawRect(xs,ys,w,h,0x07E0) ; break ;
      case 2: drawRect(xs,ys,w,h,0x001F) ; break ;
    }
    xs += 8 ;    ys += 8 ;
    w -= 16 ;    h -= 16 ;
  }
  delay(2000);

  // -- fillRect
  xs = 0 ;  ys = 0 ;  w = 128 ;  h = 128 ;
  for (int i=0;i<8;i++) {
    switch (i % 3) {
      case 0: fillRect(xs,ys,w,h,0xF800) ; break ;
      case 1: fillRect(xs,ys,w,h,0x07E0) ; break ;
      case 2: fillRect(xs,ys,w,h,0x001F) ; break ;
    }
    xs += 8 ;    ys += 8 ;
    w -= 16 ;    h -= 16 ;
  }
  delay(2000);

  // -- Circle
  fillScreen(0x0000) ; 
  xs = 64 ;  ys = 64 ;  w = 8 ;
  for (int i=0;i<8;i++) {
    switch (i % 3) {
      case 0: drawCircle(xs,ys,w,0xF800) ; break ;
      case 1: drawCircle(xs,ys,w,0x07E0) ; break ;
      case 2: drawCircle(xs,ys,w,0x001F) ; break ;
    }
    w += 8 ;
  }
  delay(2000);

  // -- FillCircle
  fillScreen(0x0000) ; 
  xs = 64 ;  ys = 64 ;  w = 64 ;
  for (int i=0;i<8;i++) {
    switch (i % 3) {
      case 0: fillCircle(xs,ys,w,0xF800) ; break ;
      case 1: fillCircle(xs,ys,w,0x07E0) ; break ;
      case 2: fillCircle(xs,ys,w,0x001F) ; break ;
    }
    w -= 8 ;
  }
  delay(3000);  
}
